unit eBob42.Planning;
{$R+}
interface
uses
  {$IFDEF CLR}
    Borland.Delphi.IniFiles
  {$ELSE}
    IniFiles
  {$ENDIF};

type
  TScore = Array of Integer;
  TName = Array of String;

  TSession = class
    Title,Speaker: String;
    ID,Room,RepeatedAtTimeslot: Integer;
    Level,Topic,OptionalSecondTopic: Integer;
    constructor Create(const IniFile: TIniFile; const Key: String;
      const Session: Integer);
  public
    SessionScore: Integer;
  protected
    function Score(const First,Second: Integer; const RoomScore: TScore; const ASpeaker: String = ''): Integer;
  end;

  TSessions = Array of TSession;

  TTimeslot = class
    Time: String;
    Sessions: Integer;
    Session: TSessions;
    Index: Array of Integer; // sorted index to sessions
    constructor Create(const IniFile: TIniFile; const Key: String);
    destructor Destroy; override;
  protected
    procedure SortSessions(const First, Second: Integer; const RoomScore: TScore; const ASpeaker: String = '');
  end;

  TTimeslots = Array of TTimeslot;

  TDCon = class
    Rooms: Integer;
    RoomScore: TScore; // room scores
    Levels: Integer;
    Level: TName; // names of Levels
    Topics: Integer;
    Topic: TName; // names of Topics
    Timeslots: Integer;
    Timeslot: TTimeslots;
    constructor Create(const IniFile: TIniFile);
    destructor Destroy; override;
  public
    procedure GeneratePlanning(const First, Second: Integer; const ASpeaker: String = '');
  end;

implementation
uses
  {$IFDEF CLR}
    Borland.Delphi.SysUtils
  {$ELSE}
    SysUtils
  {$ENDIF};

procedure Swap(var X,Y: Integer);
var
  Z: Integer;
begin
  Z := X;
  X := Y;
  Y := Z
end;

{ TSession }

constructor TSession.Create(const IniFile: TIniFile; const Key: String; const Session: Integer);
var
  S,Str: String;

  function GetTopic(Ch: Char): Integer;
  begin
    if Ord(Ch) in [Ord('1')..Ord('9')] then
      Result := Ord(Ch) - Ord('1') + 1
    else
      if Ord(Ch) in [Ord('A')..Ord('Z')] then
        Result := Ord(Ch) - Ord('A') + 10
      else
        Result := 0
  end {GetTopic};

begin
  inherited Create;
  S := IntToStr(Session);
  SessionScore := 0;
  with IniFile do
  begin
    Title := ReadString(Key, 'Title'+S, '');
    Speaker := ReadString(Key, 'Speaker'+S, '');
    ID := ReadInteger(Key, 'ID'+S, 0);
    Room := ReadInteger(Key, 'Room'+S, Session); // Room == Session
    RepeatedAtTimeslot := ReadInteger(Key, 'Repeat'+S, 0);
    Str := ReadString(Key, 'Level'+S, '0');
    if Ord(Str[1]) in [Ord('1')..Ord('3')] then
      Level := Ord(Str[1]) - Ord('1') + 1
    else Level := 0;
    Str := ReadString(Key, 'Topic'+S, '0');
    Topic := GetTopic(Str[1]);
    OptionalSecondTopic := 0;
    if Length(Str) > 1 then OptionalSecondTopic := GetTopic(Str[2])
  end
end;

function TSession.Score(const First, Second: Integer; const RoomScore: TScore; const ASpeaker: String = ''): Integer;
var
  MaxRoom: Integer;
begin
  MaxRoom := High(RoomScore);
  Result := 0;
  if Pred(Room) in [0..MaxRoom] then
    Result := Result + RoomScore[Pred(Room)];
  if (Topic = Second) or ((Level > 0) and (Level = Second)) then
    Result := Result + 2 * RoomScore[0] + 1;
  if (Topic = First) or ((Level > 0) and (Level = First)) then
    Result := Result + 4 * RoomScore[0];
  if OptionalSecondTopic <> 0 then
  begin
    if (OptionalSecondTopic = First) then Result := Result + 2 * RoomScore[0];
    if (OptionalSecondTopic = Second) then Result := Result + RoomScore[0] + 1
  end;
  if Speaker = ASpeaker then Result := Result + $10 * RoomScore[0]; // Speaker
  SessionScore := Result
end;

{ TTimeslot }

constructor TTimeslot.Create(const IniFile: TIniFile; const Key: string);
var
  i: Integer;
begin
  inherited Create;
  Time := IniFile.ReadString(Key,'time','');
  Sessions := IniFile.ReadInteger(Key,'sessions',0);
  SetLength(Index, Sessions);
  for i:=0 to Pred(Sessions) do Index[i] := i;
  SetLength(Session, Sessions);
  for i:=0 to Pred(Sessions) do
    Session[i] := TSession.Create(IniFile, Key, i+1) // start at session #1
end;

destructor TTimeslot.Destroy;
var
  i: Integer;
begin
  for i:=0 to Pred(Sessions) do Session[i].Free;
  SetLength(Session,0);
  SetLength(Index,0);
  inherited
end;

procedure TTimeslot.SortSessions(const First, Second: Integer; const RoomScore: TScore; const ASpeaker: String = '');
var
  i,j: Integer;
  Score: TScore;
begin
  SetLength(Score, Sessions);
  for i:=0 to Pred(Sessions) do
    Score[i] := Session[i].Score(First, Second, RoomScore, ASpeaker);
  for i:=0 to Pred(Pred(Sessions)) do
    for j:=Pred(Sessions) downto Succ(i) do
      if Score[Index[i]] < Score[Index[j]] then Swap(Index[i],Index[j])
end;

{ TDCon }

constructor TDCon.Create(const IniFile: TIniFile);
var
  i: Integer;
begin
  inherited Create;
  Rooms := IniFile.ReadInteger('rooms','rooms',1);
  SetLength(RoomScore, Rooms);
  for i:=0 to Pred(Rooms) do
    RoomScore[i] := IniFile.ReadInteger('rooms', Format('room%.d',[i+1]), 0);
  Levels := IniFile.ReadInteger('levels','levels',0);
  SetLength(Level, Levels);
  for i:=0 to Pred(Levels) do
    Level[i] := IniFile.ReadString('levels', Format('level%.d',[i]), '');
  Topics := IniFile.ReadInteger('topics','topics',0);
  SetLength(Topic, Topics);
  for i:=0 to Pred(Topics) do
    Topic[i] := IniFile.ReadString('topics', Format('topic%.d',[i]), '');
  Timeslots := IniFile.ReadInteger('slots','slots',0);
  SetLength(Timeslot, Timeslots);
  for i:=0 to Pred(Timeslots) do
    Timeslot[i] := TTimeslot.Create(IniFile, Format('slot%.d',[i+1]))
end;

destructor TDCon.Destroy;
var
  i: Integer;
begin
  for i:=0 to Pred(Timeslots) do Timeslot[i].Free;
  SetLength(Timeslot, 0);
  SetLength(Topic, 0);
  SetLength(Level, 0);
  SetLength(RoomScore, 0);
  inherited
end;

procedure TDCon.GeneratePlanning(const First, Second: Integer; const ASpeaker: String);
var
  i,j,k: Integer;
  SelectedSessions: Set of Byte;
begin
  for i:=0 to Pred(Timeslots) do
    Timeslot[i].SortSessions(First, Second, RoomScore, ASpeaker);

  SelectedSessions := [];
  for i:=0 to Pred(Timeslots) do
  begin
    j := Timeslot[i].Index[0];
    if (Timeslot[i].Session[j].ID in SelectedSessions) and
       (Timeslot[i].Session[j].Speaker <> ASpeaker) then
    begin
      k := Pred(Timeslot[i].Session[j].RepeatedAtTimeslot);
      if Timeslot[k].Session[Timeslot[k].Index[1]].ID in SelectedSessions then
        Swap(Timeslot[i].Index[1],Timeslot[i].Index[0]) // alternative of old one already taken
      else
      if Timeslot[i].Session[Timeslot[i].Index[1]].ID in SelectedSessions then
        Swap(Timeslot[k].Index[1],Timeslot[k].Index[0]) // alternative already taken
      else

      if (Timeslot[i].Session[Timeslot[i].Index[0]].Score(First,Second,RoomScore) +
          Timeslot[k].Session[Timeslot[k].Index[1]].Score(First,Second,RoomScore)) >
         (Timeslot[k].Session[Timeslot[k].Index[0]].Score(First,Second,RoomScore) +
          Timeslot[i].Session[Timeslot[i].Index[1]].Score(First,Second,RoomScore)) then // use previous alternative
      begin
        Swap(Timeslot[k].Index[1],Timeslot[k].Index[0]);
        SelectedSessions := SelectedSessions +
          [Timeslot[k].Session[Timeslot[k].Index[0]].ID] // Recursive check!
      end
      else // use alternative now...
      begin
        Swap(Timeslot[i].Index[1],Timeslot[i].Index[0]);
        SelectedSessions := SelectedSessions +
          [Timeslot[i].Session[Timeslot[i].Index[0]].ID] // Recursive check!
      end
    end
    else
      SelectedSessions := SelectedSessions +
        [Timeslot[i].Session[Timeslot[i].Index[0]].ID]
  end
end;

end.
 